-- gab.lua - Gabriela Avila's calculator for quantified
--   expressions, Lua version (full)
-- Author:  Eduardo Ochs <eduardoochs@gmail.com>
-- Version: 2012apr24
-- Licence: GPL3
--
-- The latest upstream version of this can be found at:
--   http://angg.twu.net/dednat5/gab.lua
--   http://angg.twu.net/dednat5/gab.lua.html
--                    (find-dn5 "gab.lua")
--
-- Older version:
--   http://angg.twu.net/dednat5/gabriela.lua
--   http://angg.twu.net/dednat5/gabriela.lua.html
--                    (find-dn5 "gabriela.lua")
--   http://angg.twu.net/dednat5/gabriela-app.lua
--   http://angg.twu.net/dednat5/gabriela-app.lua.html
--                    (find-dn5 "gabriela-app.lua")


-- Quick index:
-- .eoo.lua		(to "eoo.lua")
-- .Expr		(to "Expr")
-- .expr:infix	(to "expr:infix")
-- .constructors	(to "constructors")
-- .expr:tolisp	(to "expr:tolisp")
-- .expr:eval		(to "expr:eval")
-- .eval-quantifiers	(to "eval-quantifiers")
-- .Rect		(to "Rect")
-- .expr:print	(to "expr:print")
-- .expr:Dbg		(to "expr:Dbg")
-- .Context		(to "Context")


-- Quick index:
-- .contexts			(to "contexts")
-- .contexts-test		(to "contexts-test")
-- .def-lambda-app		(to "def-lambda-app")
-- .comprehension		(to "comprehension")
-- .comprehension-test	(to "comprehension-test")
-- .def-lambda-app-tests	(to "def-lambda-app-tests")
-- .grids			(to "grids")
-- .grids-tests		(to "grids-tests")
-- .precedence-table		(to "precedence-table")
-- .precedence		(to "precedence")
-- .recursive-descent		(to "recursive-descent")
-- .modal			(to "modal")



--                   
--   ___  ___   ___  
--  / _ \/ _ \ / _ \ 
-- |  __/ (_) | (_) |
--  \___|\___/ \___/ 
--                   
-- eoo.lua  (to ".eoo.lua")
-- dofile "eoo.lua"  -- (find-dn5 "eoo.lua")
-- Just the relevant definitions from eoo.lua, to make this self-contained.
-- The documentation is at:
--   (find-dn5 "eoo.lua" "test-eoo")
--   (find-dn5 "eoo.lua" "box-diagram")
Class = {
    type   = "Class",
    __call = function (class, o) return setmetatable(o, class) end,
  }
setmetatable(Class, Class)
otype = function (o)  -- works like type, except on my "objects"
    local  mt = getmetatable(o)
    return mt and mt.type or type(o)
  end
-- end of eoo.lua


--                                                  _   
--  _ __ ___   __ _ _ __   ___ ___  _ __   ___ __ _| |_ 
-- | '_ ` _ \ / _` | '_ \ / __/ _ \| '_ \ / __/ _` | __|
-- | | | | | | (_| | |_) | (_| (_) | | | | (_| (_| | |_ 
-- |_| |_| |_|\__,_| .__/ \___\___/|_| |_|\___\__,_|\__|
--                 |_|                                  
--
-- And here are "mapconcat" and friends, from my init file:
-- (find-angg "LUA/lua50init.lua" "map" "mapconcat")
-- (find-lua51manualw3m "#pdf-table.concat")
map = function (f, A, i, j)
    local B = {}
    for k=(i or 1),(j or #A) do table.insert(B, (f(A[k]))) end
    return B
  end
mapconcat = function (f, A, sep, i, j)
    return table.concat(map(f, A, i, j), sep)
  end
foldl = function (f, a, B, i, j)
    for k=(i or 1),(j or #B) do a = f(a, B[i]) end
    return a
  end
id  = function (...) return ... end
sorted = function (T, lt) table.sort(T, lt); return T end
keys = function (T)
    local ks = {}
    for k,_ in pairs(T) do table.insert(ks, k) end
    return ks
  end
min = function (a, b) return a < b and a or b end
max = function (a, b) return a < b and b or a end
unpack = unpack or table.unpack
-- end of mapconcat



--  _____                 
-- | ____|_  ___ __  _ __ 
-- |  _| \ \/ / '_ \| '__|
-- | |___ >  <| |_) | |   
-- |_____/_/\_\ .__/|_|   
--            |_|         
--
-- Expr  (to ".Expr")
-- expr:infix  (to ".expr:infix")
Expr = Class {
  type    = "Expr",
  __index = {
    print  = function (e) print(e); return e end,  -- uses __tostring
  },
  __tostring = function (e) return (e:infix()) end,
  __add = function (e1, e2) return Expr {[0]="+", e1, e2} end,
  __sub = function (e1, e2) return Expr {[0]="-", e1, e2} end,
  __mul = function (e1, e2) return Expr {[0]="*", e1, e2} end,
  __div = function (e1, e2) return Expr {[0]="/", e1, e2} end,
  __unm = function (e1)     return Expr {[0]="u-", e1} end,
}

-- Constructors
-- constructors  (to ".constructors")
Num  = function (n)  return Expr {[0]="n", n} end
Var  = function (s)  return Expr {[0]="v", s} end
Bool = function (b)  return Expr {[0]="b", b} end

Unm = function (e1) return - e1 end
Mul = function (e1, e2) return e1 * e2 end
Div = function (e1, e2) return e1 / e2 end
Add = function (e1, e2) return e1 + e2 end
Sub = function (e1, e2) return e1 - e2 end

Eq  = function (e1, e2) return Expr {[0]="==", e1, e2} end
Lt  = function (e1, e2) return Expr {[0]="<",  e1, e2} end
Le  = function (e1, e2) return Expr {[0]="<=", e1, e2} end
Ge  = function (e1, e2) return Expr {[0]=">=", e1, e2} end
Gt  = function (e1, e2) return Expr {[0]="<",  e1, e2} end
Neq = function (e1, e2) return Expr {[0]="!=", e1, e2} end

Not = function (e1)     return Expr {[0]="nt", e1} end
And = function (e1, e2) return Expr {[0]="&",  e1, e2} end
Or  = function (e1, e2) return Expr {[0]="|",  e1, e2} end
Imp = function (e1, e2) return Expr {[0]="->", e1, e2} end

Tuple = function (...)  return Expr {[0]="()", ...} end
Set   = function (...)  return Expr {[0]="{}", ...} end

Fa  = function (e1, e2, e3) return Expr {[0]="Fa", e1, e2, e3} end
Ex  = function (e1, e2, e3) return Expr {[0]="Ex", e1, e2, e3} end

True  = Bool(true)
False = Bool(false)

--                         ___                         _       
-- __   ____ _ _ __ ___   ( _ )     ___ ___  _ __  ___| |_ ___ 
-- \ \ / / _` | '__/ __|  / _ \/\  / __/ _ \| '_ \/ __| __/ __|
--  \ V / (_| | |  \__ \ | (_>  < | (_| (_) | | | \__ \ |_\__ \
--   \_/ \__,_|_|  |___/  \___/\/  \___\___/|_| |_|___/\__|___/
--                                                             
-- Convenience: some expressions (variables and numeric constants)
-- that I use in tests
_0 = Num(0)
_1 = Num(1)
_2 = Num(2)
_3 = Num(3)
_4 = Num(4)
_5 = Num(5)
a = Var "a"
b = Var "b"
c = Var "c"
d = Var "d"
k = Var "k"
x = Var "x"
y = Var "y"








--  _        _ _           
-- | |_ ___ | (_)___ _ __  
-- | __/ _ \| | / __| '_ \ 
-- | || (_) | | \__ \ |_) |
--  \__\___/|_|_|___/ .__/ 
--                  |_|    
--
-- expr:tolisp  (to ".expr:tolisp")
-- A Lisp-ish representation.
-- Note that our Exprs use 0-based arrays, not cons cells.
--   (find-elnode "Cons Cell Type")
--   (find-elnode "Box Diagrams")
tolisp = function (e)
    if     type(e) == "number" then return e
    elseif type(e) == "string" then return "\""..e.."\""
    elseif otype(e) == "Expr"  then
      return "("..tolisp(e[0]).." "..mapconcat(tolisp, e, " ")..")"
    else   return "<"..tostring(e)..">"
    end
  end
Expr.__index.tolisp = function (e) return tolisp(e) end
Expr.__index.lprint = function (e) print(e:tolisp()); return e end



--  _        __ _      
-- (_)_ __  / _(_)_  __
-- | | '_ \| |_| \ \/ /
-- | | | | |  _| |>  < 
-- |_|_| |_|_| |_/_/\_\
--                     

Expr.__index.infix = function (e, b)
    local op, e1, e2, e3 = e[0], e[1], e[2], e[3]
    local t, str
    if     op == "v"  then t, str = 200, e[1]
    elseif op == "n"  then t, str = 200, tostring(e[1])
    elseif op == "b"  then t, str = 200, e[1] and "T" or "F"
    elseif op == "u-" then t, str = 9, "-"..e1:infix(100)
    elseif op == "*"  then t, str = 8, e1:infix(7).."*"..e2:infix(8)
    elseif op == "/"  then t, str = 8, e1:infix(7).."*"..e2:infix(8)
    elseif op == "+"  then t, str = 7, e1:infix(6).."+"..e2:infix(7)
    elseif op == "-"  then t, str = 7, e1:infix(6).."-"..e2:infix(7)
    elseif op == "==" then t, str = 6, e1:infix(6).." == "..e2:infix(6)
    elseif op == "<=" then t, str = 6, e1:infix(6).." <= "..e2:infix(6)
    elseif op == ">=" then t, str = 6, e1:infix(6).." >= "..e2:infix(6)
    elseif op == "<"  then t, str = 6, e1:infix(6).." < " ..e2:infix(6)
    elseif op == ">"  then t, str = 6, e1:infix(6).." > " ..e2:infix(6)
    elseif op == "!=" then t, str = 6, e1:infix(6).." != "..e2:infix(6)
    elseif op == "nt" then t, str = 5, "not "..e1:infix(4)
    elseif op == "&"  then t, str = 4, e1:infix(3).." & " ..e2:infix(4)
    elseif op == "|"  then t, str = 3, e1:infix(2).." | " ..e2:infix(3)
    elseif op == "->" then t, str = 2, e1:infix(2).." -> "..e2:infix(2)
    elseif op == "()" then t, str = 200,   "("..e:infixs()..")"
    elseif op == "{}" then t, str = 200,   "{"..e:infixs().."}"
    elseif op == "Fa" then t, str = 1, "Fa "..e1:infix().." in "..
                                              e2:infix()..". "..e3:infix()
    elseif op == "Ex" then t, str = 1, "Ex "..e1:infix().." in "..
                                              e2:infix()..". "..e3:infix()
    elseif op == "Dg" then str, t = e1:infix()
    --#ifdef HAS_LAMBDA
    elseif op == "De" then t, str = 200, e1
    elseif op == "Ap" then t, str = 200, e1:infix(3).."("..e2:infix()..")"
    elseif op == "\\" then t, str = 3,   "\\"..e1:infix().."."..e2:infix()
    --#endif HAS_LAMBDA
    --#ifdef HAS_COMPREHENSION
    -- subset, set of images, generator, filter, collect
    elseif op == "Su" then t, str = 200, "{"..e1:infix().." | "..
                                              e:infixs(", ", 2, #e-1).."}"
    elseif op == "So" then t, str = 200, "{"..e[#e]:infix().." | "..
                                              e:infixs(", ", 1, #e-1).."}"
    elseif op == "<-" then t, str = 200, e1:infix().." <- "..e2:infix()
    --#endif HAS_COMPREHENSION
    --#endif HAS_MODAL
    elseif op == "Mo" then t, str = 200, e
    --#ifdef HAS_MODAL
    --                else str, t = e:infix_other(b)  -- all extensions
                      else error("Bad expr:"..tolisp(e))
    end
    if b and t <= b then str = "("..str..")" end
    return str, t
  end
Expr.__index.infixs = function (e, sep, i, j)
    return mapconcat(e.infix, e, (sep or ", "), i, j)
  end






--                  _ 
--   _____   ____ _| |
--  / _ \ \ / / _` | |
-- |  __/\ V / (_| | |
--  \___| \_/ \__,_|_|
--                    

-- expr:eval  (to ".expr:eval")
-- Evaluation.
-- To avoid making the code above too big we inject new methods into
-- Expr, using a low-level syntax for that:
--     Expr.__index.newmethod = function (e, a, b, c) ... end

etype  = function (e) return otype(e) == "Expr" and e[0] end
Expr.__index.neval = function (e)
    local ee = e:eval()
    if etype(ee) == "n" then return ee[1] end
    error("Not a number: "..tostring(ee))
  end
Expr.__index.beval = function (e)
    local ee = e:eval()
    if etype(ee) == "b" then return ee[1] end
    error("Not a boolean: "..tostring(ee))
  end
Expr.__index.seval = function (e)
    local ee = e:eval()
    if etype(ee) == "{}" then return ee end
    error("Not a set: "..tostring(ee))
  end
Expr.__index.eval_components = function (e)
    local A = map(e.eval, e)
    A[0] = e[0]
    return Expr(A)
  end

_and = function (P, Q) return P and Q end
_or  = function (P, Q) return P or  Q end
_imp = function (P, Q) return (not P) or Q end
context = {}
Expr.__index.eval = function (e)
    local op, e1, e2, e3 = e[0], e[1], e[2], e[3]
    if     op == "n" then return e
    elseif op == "b" then return e
    elseif op == "v" then return context[e1] or error("Undef: "..e1)
    elseif op == "u-" then return Num(- e1:neval())
    elseif op == "*"  then return Num(e1:neval() * e2:neval())
    elseif op == "/"  then return Num(e1:neval() / e2:neval())
    elseif op == "+"  then return Num(e1:neval() + e2:neval())
    elseif op == "-"  then return Num(e1:neval() - e2:neval())
    elseif op == "==" then return Bool(e1:neval() == e2:neval())
    elseif op == "<=" then return Bool(e1:neval() <= e2:neval())
    elseif op == ">=" then return Bool(e1:neval() >= e2:neval())
    elseif op == "<"  then return Bool(e1:neval() <  e2:neval())
    elseif op == ">"  then return Bool(e1:neval() >  e2:neval())
    elseif op == "!=" then return Bool(e1:neval() ~= e2:neval())
    elseif op == "nt" then return Bool(not e1:beval())
    elseif op == "&"  then return Bool(_and(e1:beval(), e2:beval()))
    elseif op == "|"  then return Bool(_or (e1:beval(), e2:beval()))
    elseif op == "->" then return Bool(_imp(e1:beval(), e2:beval()))
    elseif op == "{}" then return e:eval_components()
    elseif op == "()" then return e:eval_components()
    elseif op == "Fa" then return Bool(e1:_fold(true, _and, _beval, e2, e3))
    elseif op == "Ex" then return Bool(e1:_fold(false, _or, _beval, e2, e3))
    elseif op == "Dg" then return e1:DbgEval()    -- defined elsewhere
                      else return e:eval_other()  -- all extensions
    end
  end


--                          _   _  __ _               
--   __ _ _   _  __ _ _ __ | |_(_)/ _(_) ___ _ __ ___ 
--  / _` | | | |/ _` | '_ \| __| | |_| |/ _ \ '__/ __|
-- | (_| | |_| | (_| | | | | |_| |  _| |  __/ |  \__ \
--  \__, |\__,_|\__,_|_| |_|\__|_|_| |_|\___|_|  |___/
--     |_|                                            

-- eval-quantifiers  (to ".eval-quantifiers")
-- The evaluation code for "Fa" and "Ex" calls "_fold" and "_beval",
-- that are defined below.
Expr.__index.varname = function (e)
    if etype(e) == "v" then return e[1] end
    error("Not a variable: "..tostring(ee))
  end
Expr.__index.as = function (var, value, expr)
    local vname    = var:varname()
    local oldvalue = context[vname]
    context[vname] = value
    local result   = expr:eval()
    context[vname] = oldvalue
    return result
  end
Expr.__index._fold = function (var, r, op, normalize, set, expr)
    for _,value in ipairs(set) do
        r = op(r, normalize(var:as(value, expr)))
      end
    return r
  end

_beval = function (e) return e:beval() end

Expr.__index.peval  = function (e)
    local result = e:eval()
    print(tostring(e).."  -->  "..tostring(result))
    return result
  end


-- __        ___                
-- \ \      / / |__   ___ _ __  
--  \ \ /\ / /| '_ \ / _ \ '_ \ 
--   \ V  V / | | | |  __/ | | |
--    \_/\_/  |_| |_|\___|_| |_|
--                              
-- Obsolete?
Expr.__index.When = function (var, value, expr)
    return "When "  ..tostring(var)..
           "="      ..tostring(value)..
           ":  "    ..tostring(expr)..
           "  -->  "..tostring(var:as(value, expr))
  end
Expr.__index.Whens = function (var, set, expr)
    for _,value in ipairs(set) do
      print(var:When(value, expr))
    end
  end



--  ____           _   
-- |  _ \ ___  ___| |_ 
-- | |_) / _ \/ __| __|
-- |  _ <  __/ (__| |_ 
-- |_| \_\___|\___|\__|
--                     
-- Rect  (to ".Rect")
-- Rectangles (for trees)
-- (find-luamanualw3m "#pdf-string.rep")
strrep = function (str, n) return string.rep(str, max(0, n)) end
strpad = function (str, n, char) return str..strrep(char or " ", n-#str) end
torect = function (o)
    if otype(o) == "Rect" then return o end
    if otype(o) == "Expr" then return o:torect() end
    o = tostring(o)
    return Rect {w=#o, o}
  end
copy = function (A)
    local B = {}
    for k,v in pairs(A) do B[k] = v end
    setmetatable(B, getmetatable(A))
    return B
  end
Rect = Class {
  type    = "Rect",
  __index = {
    copy = function (rect) return copy(rect) end,
    get  = function (rect, y) return rect[y] or "" end,
    getp = function (rect, y, c) return strpad(rect:get(y), rect.w, c) end,
    adjw = function (rect, y) rect.w = max(rect.w, #rect:get(y)) end,
    set  = function (rect, y, str, fill)
        for i=#rect+1,y-1 do rect[i] = fill or "" end
	rect[y] = str
	rect:adjw(y)
	return rect
      end,
    lower = function (rect, n, ...)
        for i=1,(n or 1) do table.insert(rect, 1, "") end
        for y,str in ipairs({...}) do rect:set(y, str) end
        return rect
      end,
    under  = function (rect, str) return rect:copy():lower(2, str, "|") end,
    under_ = function (rect, str, n)
        return rect:under(strpad(str, rect.w+(n or 2), "_"))
      end,
  },
  __tostring = function (rect) return mapconcat(id, rect, "\n") end,
  __concat = function (r1, r2)
      r1 = torect(r1)
      r2 = torect(r2)
      r3 = Rect {w=0}
      for y=1,max(#r1, #r2) do r3:set(y, r1:getp(y, " ")..r2:get(y)) end
      return r3
    end,
}

rectconcat = function (op, rects)
    if #rects == 0 then return torect(op) end
    if #rects == 1 then return torect(rects[1]):under(op) end
    local out = torect(rects[1]):under_(op)
    for i=2,#rects-1 do out = out..torect(rects[i]):under_(".") end
    return out..torect(rects[#rects]):under(".")
  end

Expr.__index.torect = function (e)
    local op, e1 = e[0], e[1]
    if op == "n" then return torect(e1) end
    if op == "v" then return torect(e1) end
    if op == "b" then return torect(e:infix()) end
    -- local subrects = map(Expr.__index.torect, e)
    local subrects = map(torect, e)
    return rectconcat (op, subrects)
  end

-- expr:print  (to ".expr:print")
-- Some methods for printing
Expr.__index.rprint = function (e) print(e:torect()); return e end
Expr.__index.preval  = function (e)
    print(tostring(e))
    print(torect(e))
    print()
    local result = e:eval()
    print("  -->  "..tostring(result))
    print()
    return result
  end



--  ____  _           
-- |  _ \| |__   __ _ 
-- | | | | '_ \ / _` |
-- | |_| | |_) | (_| |
-- |____/|_.__/ \__, |
--              |___/ 
--
-- expr:Dbg  (to ".expr:Dbg")
-- Support for debugging (verbose subexpressions)
Expr.__index.Dbg     = function (e1) return Expr {[0]="Dg", e1} end
Expr.__index.DbgEval = function (expr)
    local result = expr:eval()
    print(contextstring()..tostring(expr).."  -->  "..tostring(result))
    return result
  end
contextstring = function ()
    local sk = sorted(keys(context))
    local f = function (name) return name.."="..tostring(context[name]) end
    return "["..mapconcat(f, sk, " ").."]  "
  end




--   ____            _            _   
--  / ___|___  _ __ | |_ _____  _| |_ 
-- | |   / _ \| '_ \| __/ _ \ \/ / __|
-- | |__| (_) | | | | ||  __/>  <| |_ 
--  \____\___/|_| |_|\__\___/_/\_\\__|
--                                    
-- Context  (to ".Context")
-- Contexts are used for variables and for substitutions.
Context = Class {
  type    = "Context",
  __index = {
    push  = function (ctxt, name, expr)
        local pair = {name, ctxt[name]}
	table.insert(ctxt.__stack, pair)
	ctxt[name] = expr
        return ctxt
      end,
    pop   = function (ctxt)
        local pair = table.remove(ctxt.__stack) or error("Empty __stack")
	local name, oldvalue = pair[1], pair[2]
	ctxt[name] = oldvalue
	return ctxt
      end,
    minus = function (ctxt, name)
        if ctxt[name] == nil then return ctxt end
        ctxt = copy(ctxt); ctxt[name] = nil; return ctxt
      end,
    keys     = function (ctxt) return sorted(keys(ctxt:minus("__stack"))) end,
    tostring = function (ctxt, sep)
        local ks = ctxt:keys()
        for i,name in ipairs(ks) do ks[i] = name.."="..tostring(ctxt[name]) end
        return mapconcat(id, ks, sep or ", ")
      end,
    print  = function (ctxt) print(ctxt); return ctxt end,
    vprint = function (ctxt) print(ctxt:tostring"\n"); return ctxt end,
  },
  __tostring = function (ctxt) return ctxt:tostring() end,
}

newcontext = function (T) T.__stack = {}; return Context(T) end

context = newcontext {}
defs    = newcontext {}




--  _                 _         _       
-- | | __ _ _ __ ___ | |__   __| | __ _ 
-- | |/ _` | '_ ` _ \| '_ \ / _` |/ _` |
-- | | (_| | | | | | | |_) | (_| | (_| |
-- |_|\__,_|_| |_| |_|_.__/ \__,_|\__,_|
--                                      
-- def-lambda-app  (to ".def-lambda-app")
-- A very simple implementation of definitions and application
Def    = function (name)      return Expr {[0]="De", name} end
App    = function (e1, e2)    return Expr {[0]="Ap", e1, e2} end
Lambda = function (var, expr) return Expr {[0]="\\", var, expr} end

Expr.__index.eval_other = function (e)
    local op, e1, e2, e3 = e[0], e[1], e[2], e[3]
    if     op == "De" then return defs[e1]:eval()
    elseif op == "\\" then return e
    elseif op == "Ap" then return _app(e1:eval(), e2:eval())
    elseif op == "Su" then return _evalcompr(e)
    elseif op == "So" then return _evalcompr(e)
                      else print(tolisp(e)); error("Bad expr")
    end
  end

_app = function (f, a)
    if otype(f) ~= "Expr" then error("f must be an expr") end
    if otype(a) ~= "Expr" then error("a must be an expr") end
    if f[1][0]  ~= "v"    then error("x must ber a var")  end
    local f0, fvarname, fexpr = f[0], f[1][1], f[2]
    if f0 ~= "\\" then error("f must be a lambda form") end
    context:push(fvarname, a)
    local result = fexpr:eval()
    context:pop()
    return result
  end



--   ____                               _                    _             
--  / ___|___  _ __ ___  _ __  _ __ ___| |__   ___ _ __  ___(_) ___  _ __  
-- | |   / _ \| '_ ` _ \| '_ \| '__/ _ \ '_ \ / _ \ '_ \/ __| |/ _ \| '_ \ 
-- | |__| (_) | | | | | | |_) | | |  __/ | | |  __/ | | \__ \ | (_) | | | |
--  \____\___/|_| |_| |_| .__/|_|  \___|_| |_|\___|_| |_|___/_|\___/|_| |_|
--                      |_|                                                

-- comprehension  (to ".comprehension")
-- (find-dn5 "gab-tests.lua" "comprehension-1")
-- Inspired by:
--   (find-es "haskell" "comprehension")
--   http://www.haskell.org/tutorial/goodies.html#tut-list-comps
--   http://www.haskell.org/onlinereport/exps.html#list-comprehensions
--   http://www.haskell.org/haskellwiki/List_comprehension
--   http://en.wikipedia.org/wiki/List_comprehension

Subset  = function (...)      return Expr {[0]="Su", ...} end
Setof   = function (...)      return Expr {[0]="So", ...} end
Gen     = function (var, set) return Expr {[0]="<-", var, set} end
-- Filt    = function (expr)     return Expr {[0]="Cf", expr} end
-- Collect = function (expr)     return Expr {[0]="Co", expr} end

_evalcompr_ = function (e, i)
    local op, e1, e2 = e[i][0], e[i][1], e[i][2]
    if     op == "Co" then _collect(e1:eval())
    elseif op == "Cf" then
      if not e1:beval() then return end
      _evalcompr_(e, i+1)
    elseif op == "<-" then
      local varname = e1[1]
      local set     = e2:seval()
      -- print(tolisp(varname), tolisp(e2), tolisp(set))
      for _,value in ipairs(set) do
        context:push(varname, value)
        _evalcompr_(e, i+1)
        context:pop()
      end
    else print(i, op, e1, e2); error("what?")
    end
  end

_evalcompr_ = function (es, collect, i)
    local e, op, e1, e2 = es[i], es[i][0], es[i][1], es[i][2]
    if i == #es then collect(e:eval()); return end
    if op == "<-" then                  -- generator
      if e1[0] ~= "v" then error("Not a var") end
      local varname = e1[1]
      local set     = e2:seval()
      for _,value in ipairs(set) do
        context:push(varname, value)
        _evalcompr_(es, collect, i+1)
        context:pop()
      end
    else                                -- filter
      if not e:beval() then return end  --  on F: abort
      _evalcompr_(es, collect, i+1)     --  on T: continue
    end
  end

_evalcompr = function (e)
    local results = Set()
    local collect = function (e) table.insert(results, e) end
    _evalcompr_(e, collect, 1)
    return results
  end




--             _     _     
--   __ _ _ __(_) __| |___ 
--  / _` | '__| |/ _` / __|
-- | (_| | |  | | (_| \__ \
--  \__, |_|  |_|\__,_|___/
--  |___/                  
--
-- grids  (to ".grids")
-- Some functions to create tables.
-- Too many things are hardcoded at the moment.
--
withdef = function (name, def, expr)
    defs:push(name, def)
    local result = expr:eval()
    defs:pop()
    return result
  end

cell = function (x, y)
    if y == 1 then return Cols[x] end
    if x == 1 then return Ps[y-1] end
    return withdef("P", Lambda("k", Ps[y-1]), Cols[x])
  end
column = function (x)
    local C = Rect {w=0}
    for y=1,#Ps+1 do C:set(y, cell(x, y):infix()) end
    return C
  end
columns = function (x1, x2)
    x1 = x1 or 1
    x2 = x2 or #Cols
    local result = column(x1)
    for x=x1+1,x2 do result = result.."  "..column(x) end
    return result
  end

efold = function (f, a, ...)
    local result = a
    for _,b in ipairs {...} do result = f(result, b) end
    return result
  end
Ors  = function (...) return efold(Or,  ...) end
Ands = function (...) return efold(And, ...) end

P = Def "P"

P1, P2, P3, P4 = App(P, _1), App(P, _2), App(P, _3), App(P, _4)
defs.E_1 = Ors (    P2, P3    )
defs.E_2 = Ors (P1, P2, P3, P4)
defs.E_3 = Ands(    P2, P3    )
defs.E_4 = Ands(P1, P2, P3, P4)
defs["E'_1"] = Or (Ors (P1, P2, P3), Ors (P2, P3, P4))
defs["E'_2"] = And(Ors (P1, P2, P3), Ors (P2, P3, P4))
defs["E'_3"] = Or (Ands(P1, P2, P3), Ands(P2, P3, P4))
defs["E'_4"] = And(Ands(P1, P2, P3), Ands(P2, P3, P4))

P123or   = Ors (P1, P2, P3)
P123and  = Ands(P1, P2, P3)
P234or   = Ors (P2, P3, P4)
P234and  = Ands(P2, P3, P4)
P123or   = Ex(k, Set(_1, _2, _3), App(P, k))
P123and  = Fa(k, Set(_1, _2, _3), App(P, k))
P234or   = Ex(k, Set(_2, _3, _4), App(P, k))
P234and  = Fa(k, Set(_2, _3, _4), App(P, k))
defs["E'_1"] = Or (P123or,  P234or)
defs["E'_2"] = And(P123or,  P234or)
defs["E'_3"] = Or (P123and, P234and)
defs["E'_4"] = And(P123and, P234and)

Cols = {
  App(P,  k),
  P1, P2, P3, P4,
  Def("E_1"), Def("E_2"), Def("E_3"), Def("E_4"),
}
Ps = {
   Ge(    k, _2),
   Ge( _2*k, _2),
   Eq(    k, _1),
   Lt(    k, _1),
  Neq(    k, _2),
  Neq(    k, _3),
  And(Neq(k, _2), Neq(k, _3)),
  Neq(    k, _0),
}




--                            
--  _ __   __ _ _ __ ___  ___ 
-- | '_ \ / _` | '__/ __|/ _ \
-- | |_) | (_| | |  \__ \  __/
-- | .__/ \__,_|_|  |___/\___|
-- |_|                        

-- precedence-table  (to ".precedence-table")
-- precedence  (to ".precedence")
-- (find-dn5 "gab-tests.lua" "parse-2")
-- (find-dn5 "gab-tests.lua" "parse-3")
ops = {}
binop = function (op, top, l, r)
    ops[op] = {kind="binop", top=top, l=l, r=r}
  end
prefix = function (op, top, r)
    ops[op] = {kind="prefix", top=top, r=r}
  end
binop("*",  8, 7, 8)
binop("/",  8, 7, 8)
binop("+",  7, 6, 7)
binop("-",  7, 6, 7)
binop("==", 6, 6, 6)
binop("<=", 6, 6, 6)
binop(">=", 6, 6, 6)
binop("<",  6, 6, 6)
binop(">",  6, 6, 6)
binop("!=", 6, 6, 6)
binop("&",  4, 3, 4)
binop("|",  3, 2, 3)
binop("->", 2, 2, 2)
prefix("u-", 9, 100)
prefix("nt", 5, 4)

binop("in", 6, 6, 6)
binop("<-", 6, 6, 6)

-- recursive-descent  (to ".recursive-descent")
subj = split "2 + 3 * 4 + 5 * 6 * 7 (eof)"
pos  = 1
parse = function (str, b)
    subj = split(str)       -- token are separated by spaces
    pos = 1
    return pa_expr(b)
  end
pparse = function (str, b)  -- print the result of a parse
    local e = parse(str, b)
    print(e:tolisp())
    print(e:torect())
    return e
  end
pparsee = function (str, b) pparse(str):peval() end

-- Abbreviations:
--   "la"  = "looking at"
--   "pa"  = "parse"
--   "wos" = "without suffixes"

-- We use a very simple recursive parser with one token of lookahead.
-- It operates on the array "subj", that at the moment is just an
-- array of strings obtained by running "split" on a given string -
-- to avoid lexing. Our grammar is:
--
--   parenexpr   ::= ( expr )
--   setexpr     ::= { }
--                 | { expr commaexpr^* }
--                 | { expr | expr commaexpr^* }
--   qlexpr      ::= Fa expr . expr
--                 | Ex expr . expr
--                 | \\ expr . expr
--   exprwos     ::= parenexpr
--                 | setexpr
--                 | qlexpr
--                 | prefixop expr
--                 | number
--                 | var
--   complement1 ::= suffixop
--                 | parenexpr
--                 | binop expr
--   expr        ::= exprwos complement1^*
--
-- Note that each alternative starts with a different token - we use
-- the "la_*" functions to choose which alternative to follow.

perror = function (str) error("(la = `"..la().."') "..str, 2) end

la  = function () return subj[pos] end
eat = function () pos = pos+1; return subj[pos-1] end
pa_forced = function (token)
    if la() ~= token then perror("Expected `"..token.."'") end
    return eat()
  end

under       = function (b, op) return (not b) or b <= ops[op].top end
la_op       = function (k) return ops[la()] and ops[la()].kind == k end
la_op_under = function (k, b) return la_op(k) and under(b, la()) end

la_binop  = function (b) return la_op_under("binop",  b) end
la_prefix = function (b) return la_op_under("prefix", b) end
la_suffix = function (b) return la_op_under("suffix", b) end

-- 

la_number = function () return (la():match"^[0-9]+$") end
pa_number = function () return Num(eat()+0) end

la_var = function () return (la():match"^[A-Za-z]+$") end
pa_var = function () return Var(eat()) end

la_paren      = function () return la() == "(" end
pa_parenclose = function () return pa_forced(")") end
pa_parenexpr  = function ()
    local o, e, c = eat(), pa_expr(), pa_parenclose()
    return e
  end
pa_parenexpr  = function ()
    local o, e, es, c = eat(), pa_expr(), pa_commaexprs(), pa_parenclose()
    if #es == 0 then return e else return Tuple(e, unpack(es)) end
  end

la_expr = function ()
    return la_number() or la_var() or la_paren() or la_set() or la_prefix()
  end
pa_exprwos = function ()
    -- PP("pa_exprwos with pos = "..pos)
    if     la_paren()  then return pa_parenexpr()
    elseif la_set()    then return pa_set()
    elseif la_ql()     then return pa_qlexpr()
    elseif la_prefix() then
      local op = eat()
      local e  = pa_expr(ops[op].r)
      return Expr {[0]=op, e}
    elseif la_number() then return pa_number()
    elseif la_var()    then return pa_var()
    end
    perror("Not an expr (in exprwos)")
  end

la_complement = function (b)
    return la_suffix(b) or la_paren() or la_binop(b)
  end
pa_expr = function (b)
    -- PP("pa_expr with pos = "..pos)
    local e = pa_exprwos()
    while la_complement(b) do
      if     la_suffix(b) then e = Expr {[0]=eat(), e}
      elseif la_paren()   then e = App(e, pa_parenexpr())
      elseif la_binop(b)  then
        local op = eat()
        local re = pa_expr(ops[op].r)
        e = Expr {[0]=op, e, re}
      else perror("Not an expr")
      end
    end
    return e
  end

-- (find-angg "LUA/lua50init.lua" "pack-and-unpack")
la_set      = function () return la() == "{" end
pa_setclose = function () return pa_forced("}") end
pa_commaexprs = function ()
    local es = {}
    while la() == "," do eat(); table.insert(es, pa_expr()) end
    return es
  end
pa_set = function ()
    eat()
    if     la() == "}" then eat(); return Set()
    elseif la_expr() then
      local e = pa_expr(4)
      if la() == "|" then
        local _, er, ers, _ = eat(), pa_expr(), pa_commaexprs(), pa_setclose()
        if e[0] == "<-" then
          local s = Subset(e, er, unpack(ers))
          table.insert(s, e[1])
          return s
        else 
          local s = Setof(er, unpack(ers))
          table.insert(s, e)
          return s
        end
      else local es, _ = pa_commaexprs(), pa_setclose()
        return Set(e, unpack(es))
      end
    end
    perror("Not a set")
  end

la_ql     = function () return la() == "Fa" or la() == "Ex" or la() == "\\" end
pa_dot    = function () return pa_forced(".") end
pa_qlexpr = function ()
    local ql, vare, dot, bodye = eat(), pa_expr(), pa_dot(), pa_expr()
    if ql == "\\" then return Lambda(vare, bodye) end
    if vare[0] == "<-" or vare[0] == "in" then
      return Expr {[0]=ql, vare[1], vare[2], bodye}
    else
      perror("varexpr must be `_ <- _' or `_ in _'")
    end 
  end




--  __  __           _       _ 
-- |  \/  | ___   __| | __ _| |
-- | |\/| |/ _ \ / _` |/ _` | |
-- | |  | | (_) | (_| | (_| | |
-- |_|  |_|\___/ \__,_|\__,_|_|
--                             
-- modal  (to ".modal")
-- (find-angg "LUA/canvas3.lua")
-- (find-angg "LUA/canvas3.lua" "ZDag-functions")

-- It should be easy to make the logic of the calculator more
-- flexible... at present it is boolean, two-valued, with the same
-- truth values as the underlying Lua.

Vee_ = function (s)
    local a, b, c = s:sub(1,1), s:sub(2,2), s:sub(3,3)
    return Rect {w=3, a.." "..b, " "..c}
  end
Vee = function (s)
    return Expr {[0]="Mo", Vee_(s)}
  end




-- This is just a library.
print("Loading: gab.lua")
io.stdout:setvbuf("no")    -- (find-es "lua5" "setvbuf")






-- Local Variables:
-- coding:             raw-text-unix
-- ee-anchor-format:   "%s"
-- End:
